import base64
from io import BytesIO

import qrcode


def _pil_to_base64(pil_img):
    output = BytesIO()
    pil_img.save(output, format="PNG")
    img = output.getvalue()

    # encode to base64
    img = base64.b64encode(img)
    return img.decode()


def get_qr_code_image_base64(message):
    qr = qrcode.QRCode(
        version=1,
        error_correction=qrcode.constants.ERROR_CORRECT_L,
    )
    qr.add_data(message)
    qr.make(fit=True)
    img = qr.make_image(fill_color="black", back_color="white")
    return _pil_to_base64(img)


def get_qr_code_image_html(message, pixels=250):
    base64 = get_qr_code_image_base64(message)

    pixels = int(pixels)
    html = f'<img src="data:image/png;base64,{base64}" style="width: {pixels}px; height: auto">'
    return html



from enum import Enum, unique


@unique
class States(Enum):
    # from https://einvoice1.gst.gov.in/Documents/EINVOICE_SCHEMA.xlsx
    JAMMU_AND_KASMIR = "1"
    HIMACHAL_PRADESH = "2"
    PUNJAB = "3"
    CHANDIGARH = "4"
    UTTARAKHAND = "5"
    HARYANA = "6"
    DELHI = "7"
    RAJASTHAN = "8"
    UTTAR_PRADESH = "9"
    BIHAR = "10"
    SIKKIM = "11"
    ARUNACHAL_PRADESH = "12"
    NAGALAND = "13"
    MANIPUR = "14"
    MIZORAM = "15"
    TRIPURA = "16"
    MEGHALAYA = "17"
    ASSAM = "18"
    WEST_BENGAL = "19"
    JHARKHAND = "20"
    ORISSA = "21"
    CHHATTISGARH = "22"
    MADHYA_PRADESH = "23"
    GUJARAT = "24"
    DAMAN_AND_DIU = "25"
    DADAR_AND_NAGAR_HAVELI = "26"
    MAHARASTRA = "27"
    KARNATAKA = "29"
    GOA = "30"
    LAKSHADWEEP = "31"
    KERALA = "32"
    TAMIL_NADU = "33"
    PONDICHERRY = "34"
    ANDAMAN_AND_NICOBAR = "35"
    TELENGANA = "36"
    ANDHRA_PRADESH = "37"
    LADAKH = "38"
    OTHER_TERRITORY = "97"
    OTHER_COUNTRIES = "96"
    CENTRE_JURISDICTION = "99"



from .generators import get_buyer_dtls


def to_buyer(gst_info, place_of_supply):
    parts = [
        gst_info["AddrBnm"],
        gst_info["AddrBno"],
        gst_info["AddrFlno"],
        gst_info["AddrSt"],
    ]
    addr1 = ", ".join(part.strip() for part in parts if part.strip())
    return get_buyer_dtls(
        gstin=gst_info["Gstin"],
        lgl_nm=gst_info["LegalName"],
        addr1=addr1,
        pos=place_of_supply,
        loc=gst_info["AddrLoc"],
        pin=gst_info["AddrPncd"],
        stcd=str(gst_info["StateCode"]),
    )



import base64

from cryptography.hazmat.primitives import padding as sym_padding
from cryptography.hazmat.primitives.asymmetric import padding as asym_padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.serialization import load_pem_public_key

# specs for algorithms are available at:
# https://einv-apisandbox.nic.in/FaqsonAPI.html


def encrypt_with_rsa_pub_key(message, public_key_str) -> str:
    """
    encrypt message with RSA by given public key
    """
    public_key = load_pem_public_key(public_key_str.encode())
    encrypted_msg = public_key.encrypt(
        plaintext=message, padding=asym_padding.PKCS1v15()
    )
    encoded_encrypted_msg = base64.b64encode(encrypted_msg).decode()
    return encoded_encrypted_msg


def decrypt_with_aes(message, key, raw=False):
    """
    decrypts the message using the given secret key (SEK)
    """
    message = base64.b64decode(message)
    key = base64.b64decode(key)
    decryptor = Cipher(
        algorithm=algorithms.AES(key=key), mode=modes.ECB()
    ).decryptor()
    decrypted = decryptor.update(message) + decryptor.finalize()

    # remove padding with pkcs7
    unpadder = sym_padding.PKCS7(128).unpadder()
    decrypted = unpadder.update(decrypted) + unpadder.finalize()

    # convert to string if not raw
    if not raw:
        decrypted = decrypted.decode()
    return decrypted


def encrypt_with_aes(message, key) -> str:
    """
    encrypts the message with the given secret key (SEK)
    """
    key = base64.b64decode(key)
    encryptor = Cipher(
        algorithm=algorithms.AES(key=key), mode=modes.ECB()
    ).encryptor()

    # pad the message with PKCS7
    padder = sym_padding.PKCS7(128).padder()
    message = padder.update(message) + padder.finalize()

    # encrypt the message
    message = encryptor.update(message)

    # base64 encode the message
    message = base64.b64encode(message).decode()
    return message



import base64
import datetime as dt
import json
import logging
import os
from hashlib import md5
from pathlib import Path
from pprint import pformat

import requests

from . import crypto

logger = logging.getLogger(__name__)


def _get_decrypted_sek(sek, app_key):
    decrypted_sek = crypto.decrypt_with_aes(sek, app_key, raw=True)

    # handle base64 encoding
    decrypted_sek = base64.b64encode(decrypted_sek).decode()
    return decrypted_sek


def _get_app_key() -> str:
    """
    creates a random app_key of 32 byte
    """
    aes_key = os.urandom(32)
    readable = base64.b64encode(aes_key).decode()
    return readable


class RequestError(Exception):
    pass


class GenerateTokenError(Exception):
    pass


def _raise_formatted_error(err, msg):
    fmt_err = pformat(err, indent=4)
    logger.error(fmt_err)
    raise RequestError(msg, err)


def _get_data_from_response(response, *, encryption_key):
    if response.status_code == 200:
        response = response.json()
        if response["Status"] == 1:
            data = response["Data"]
            if encryption_key:
                data = crypto.decrypt_with_aes(data, encryption_key)
                data = json.loads(data)
            return data
        else:
            _raise_formatted_error(response, "action failed")
    else:
        _raise_formatted_error(response.text, f"status {response.status_code}")


def _get_auth_token(session, force_regenerate_token):
    app_key = _get_app_key()
    url = f"{session.base_url}/eivital/v1.04/auth"
    # request headers
    headers = {
        "client-id": session.client_id,
        "client-secret": session.client_secret,
        "gstin": session.gstin,
    }
    if session.gsp_headers:
        headers.update(session.gsp_headers)

    # request payload
    payload = {
        "UserName": session.username,
        "Password": session.password,
        "AppKey": app_key,
        "ForceRefreshAccessToken": force_regenerate_token,
    }
    # convert payload to json string
    payload = json.dumps(payload)

    # base64 encoding of payload
    payload = base64.b64encode(payload.encode())

    # encrypt using e-Invoice public Key
    payload = crypto.encrypt_with_rsa_pub_key(payload, session.public_key)

    payload = {"Data": payload}
    response = requests.post(url, json=payload, headers=headers)

    data = _get_data_from_response(response, encryption_key=None)
    return data, app_key


def _get_store_path(cache_dir, session):
    # prepare dictionary of all session variables
    attr_dict = {
        a: getattr(session, a)
        for a in dir(session)
        if not a.startswith("_") and not callable(getattr(session, a))
    }

    # create a hash for session values
    repr = json.dumps(attr_dict, sort_keys=True)
    cache_key = md5(repr.encode()).hexdigest()

    filename = f"{cache_key}.json"
    store_path = Path(cache_dir) / filename
    return store_path


def _get_stored_token(cache_dir, session):
    store_path = _get_store_path(cache_dir, session)
    try:
        stored_token = json.loads(store_path.read_text())

        # check expiry
        expiry = dt.datetime.strptime(
            stored_token["expiry_time"], "%Y-%m-%d %H:%M:%S"
        )
        if expiry < (dt.datetime.now() + dt.timedelta(minutes=10)):
            stored_token = None

        # return tuple of token and sek
        else:
            stored_token = (stored_token["token"], stored_token["sek"])
    except Exception:
        stored_token = None
    return stored_token


def _store_token(cache_dir, session, expiry_time):
    store_path = _get_store_path(cache_dir, session)
    try:
        store_path.parent.mkdir(parents=True, exist_ok=True)
        obj = {
            "token": session._auth_token,
            "sek": session._auth_sek,
            "expiry_time": expiry_time,
        }
        store_path.write_text(json.dumps(obj))
    except Exception:
        pass


class Session:
    def __init__(
        self,
        gstin,
        client_id,
        client_secret,
        username,
        password,
        public_key,
        base_url="https://einv-apisandbox.nic.in",
        gsp_headers=None,
    ):
        self.gstin = gstin
        self.client_id = client_id
        self.client_secret = client_secret
        self.username = username
        self.password = password
        self.public_key = public_key

        self.base_url = base_url
        self.gsp_headers = gsp_headers

        self._auth_token = None
        self._auth_sek = None

    def generate_token(
        self, force_regenerate_token=False, cache_dir="tokens_cache"
    ):
        stored_token = _get_stored_token(cache_dir, self)
        if stored_token and not force_regenerate_token:
            self._auth_token, self._auth_sek = stored_token
        else:
            data, app_key = _get_auth_token(self, force_regenerate_token)
            self._auth_token = data["AuthToken"]
            self._auth_sek = _get_decrypted_sek(data["Sek"], app_key)

            # cache token
            expiry_time = data["TokenExpiry"]
            _store_token(cache_dir, self, expiry_time)

    def _get_request_headers(self):
        headers = {
            "client-id": self.client_id,
            "client-secret": self.client_secret,
            "gstin": self.gstin,
            "user_name": self.username,
            "AuthToken": self._auth_token,
        }
        if self.gsp_headers:
            headers.update(self.gsp_headers)
        return headers

    def get(self, url, headers_extra=None):
        """
        sends get request to the given url along with authentication headers
        returns decrypted response
        """
        if not self._auth_sek:
            raise GenerateTokenError()

        headers = self._get_request_headers()
        if headers_extra:
            headers.update(headers_extra)

        response = requests.get(url, headers=headers)
        return _get_data_from_response(response, encryption_key=self._auth_sek)

    def post(self, url, data, headers_extra=None):
        """
        sends post request to the given url
        - adds authentication headers
        - encrypts the payload

        returns decrypted response
        """
        if not self._auth_sek:
            raise GenerateTokenError()

        # convert payload to json string
        data = json.dumps(data)
        payload = data.encode()

        # encrypt payload
        payload = crypto.encrypt_with_aes(payload, self._auth_sek)
        payload = {"Data": payload}
        headers = self._get_request_headers()
        if headers_extra:
            headers.update(headers_extra)

        response = requests.post(url, json=payload, headers=headers)
        return _get_data_from_response(response, encryption_key=self._auth_sek)

    def get_gst_info(self, party_gstin):
        """
        fetches and returns info for the given gst number
        """
        url = f"{self.base_url}/eivital/v1.04/Master/gstin/{party_gstin}"
        return self.get(url)

    def generate_e_invoice(self, invoice):
        """
        generates and returns e-invoice for given invoice

        e-invoice contains IRN, QR-code and signature
        """
        url = f"{self.base_url}/eicore/v1.03/Invoice"
        return self.post(url, invoice)

    def get_e_invoice_by_irn(self, irn, sup_gstin=None):
        """
        returns e-invoice for an already generated irn
        """
        url = f"{self.base_url}/eicore/v1.03/Invoice/irn/{irn}"
        return self.get(url)



def _to_camel_case(name):
    return "".join(word.title() for word in name.split("_"))


def _to_camel_case_dict(**kwargs):
    return {_to_camel_case(k): v for k, v in kwargs.items()}


def get_tran_dtls(*, tax_sch="GST", sup_typ="B2B", **kwargs):
    return _to_camel_case_dict(tax_sch=tax_sch, sup_typ=sup_typ, **kwargs)


def get_doc_dtls(*, typ, no, dt, **kwargs):
    return _to_camel_case_dict(typ=typ, no=no, dt=dt, **kwargs)


def get_seller_dtls(*, gstin, lgl_nm, addr1, loc, pin, stcd, **kwargs):
    return _to_camel_case_dict(
        gstin=gstin,
        lgl_nm=lgl_nm,
        addr1=addr1,
        loc=loc,
        pin=pin,
        stcd=stcd,
        **kwargs,
    )


def get_buyer_dtls(*, gstin, lgl_nm, pos, addr1, loc, pin, stcd, **kwargs):
    return _to_camel_case_dict(
        gstin=gstin,
        lgl_nm=lgl_nm,
        pos=pos,
        addr1=addr1,
        loc=loc,
        pin=pin,
        stcd=stcd,
        **kwargs,
    )


def get_disp_dtls(*, nm, addr1, loc, pin, stcd, **kwargs):
    return _to_camel_case_dict(
        nm=nm, addr1=addr1, loc=loc, pin=pin, stcd=stcd, **kwargs
    )


def get_ship_dtls(*, lgl_nm, addr1, loc, pin, stcd, **kwargs):
    return _to_camel_case_dict(
        lgl_nm=lgl_nm, addr1=addr1, loc=loc, pin=pin, stcd=stcd, **kwargs
    )


def get_item(
    *,
    sl_no,
    is_servc,
    hsn_cd,
    unit_price,
    igst_amt,
    tot_amt,
    ass_amt,
    gst_rt,
    tot_item_val,
    **kwargs,
):
    return _to_camel_case_dict(
        sl_no=sl_no,
        is_servc=is_servc,
        hsn_cd=hsn_cd,
        unit_price=unit_price,
        igst_amt=igst_amt,
        tot_amt=tot_amt,
        ass_amt=ass_amt,
        gst_rt=gst_rt,
        tot_item_val=tot_item_val,
        **kwargs,
    )


def get_bch_dtls(*, nm, **kwargs):
    return _to_camel_case_dict(nm=nm, **kwargs)


def get_val_dtls(*, tot_inv_val, **kwargs):
    return _to_camel_case_dict(tot_inv_val=tot_inv_val, **kwargs)


def get_pay_dtls(**kwargs):
    return _to_camel_case_dict(**kwargs)


def get_ewb_dtls(*, distance, **kwargs):
    return _to_camel_case_dict(distance=distance, **kwargs)


def get_invoice(
    *,
    version="1.1",
    tran_dtls,
    doc_dtls,
    seller_dtls,
    buyer_dtls,
    item_list,
    val_dtls,
    **kwargs,
):
    return _to_camel_case_dict(
        version=version,
        tran_dtls=tran_dtls,
        doc_dtls=doc_dtls,
        seller_dtls=seller_dtls,
        buyer_dtls=buyer_dtls,
        item_list=item_list,
        val_dtls=val_dtls,
        **kwargs,
    )



from .generators import (
    get_bch_dtls,
    get_buyer_dtls,
    get_disp_dtls,
    get_doc_dtls,
    get_ewb_dtls,
    get_invoice,
    get_item,
    get_pay_dtls,
    get_seller_dtls,
    get_ship_dtls,
    get_tran_dtls,
    get_val_dtls,
)
from .session import Session
